#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
from khiva.linalg import *
from khiva.array import *
from khiva.library import set_backend, KHIVABackend


########################################################################################################################

class LinalgTest(unittest.TestCase):
    DELTA = 1e-6

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_lls(self):
        lls_result = lls(Array(np.array([[4, 3], [-1, -2]])), Array([3, 1], dtype.f32))
        a = lls_result.to_numpy()
        self.assertAlmostEqual(a[0], 1, delta=self.DELTA)
        self.assertAlmostEqual(a[1], 1, delta=self.DELTA)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(LinalgTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################


def covariance(tss, unbiased=False):
    """ Returns the covariance matrix of the time series contained in tss.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series.
    :param unbiased: Determines whether it divides by n - 1 (if false) or n (if true).
    :return: The covariance matrix of the time series.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.covariance_statistics(ctypes.pointer(tss.arr_reference),
                                                         ctypes.pointer(ctypes.c_bool(unbiased)),
                                                         ctypes.pointer(b))
    return Array(array_reference=b)


def kurtosis(tss):
    """ Returns the kurtosis of tss (calculated with the adjusted Fisher-Pearson standardized moment coefficient G2).

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series.
    :return: The kurtosis of tss.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.kurtosis_statistics(ctypes.pointer(tss.arr_reference),
                                                       ctypes.pointer(b))
    return Array(array_reference=b)


def ljung_box(tss, lags):
    """ The Ljung–Box test checks that data whithin the time series are independently distributed (i.e. the correlations
    in the population from which the sample is taken are 0, so that any observed correlations in the data result from
    randomness of the sampling process). Data are no independently distributed, if they exhibit serial correlation.
    The test statistic is:

    .. math::
        Q = n\\left(n+2\\right)\sum_{k=1}^h\\frac{\hat{\\rho}^2_k}{n-k}

    where :math:`n` is the sample size, :math:`\hat{\\rho}k` is the sample autocorrelation at lag :math:`k`, and
    :math:`h` is the number of lags being tested. Under :math:`H_0` the statistic Q follows a :math:`\\chi^2{(h)}`.
    For significance level :math:`\\alpha`, the :math:`critical region` for rejection of the hypothesis of randomness
    is:

    .. math::
        Q > \\chi_{1-\\alpha,h}^2

    where :math:`\\chi_{1-\\alpha,h}^2` is the :math:`\\alpha`-quantile of the chi-squared distribution with :math:`h`
    degrees of freedom.

    [1] G. M. Ljung G. E. P. Box (1978). On a measure of lack of fit in time series models.
    Biometrika, Volume 65, Issue 2, 1 August 1978, Pages 297–303.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
                dimension one indicates the number of time series.
    :param lags: Number of lags being tested.
    :return: Array containing the Ljung-Box statistic test.
    """
    ljung_box_out = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.ljung_box(ctypes.pointer(tss.arr_reference), ctypes.pointer(ctypes.c_long(lags)),
                                             ctypes.pointer(ljung_box_out))
    return Array(array_reference=ljung_box_out)


def moment(tss, k):
    """ Returns the kth moment of the given time series.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series.
    :param k: The specific moment to be calculated.
    :return: The kth moment of the given time series.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.moment_statistics(ctypes.pointer(tss.arr_reference),
                                                     ctypes.pointer(ctypes.c_int(k)),
                                                     ctypes.pointer(b))
    return Array(array_reference=b)


def quantile(tss, q, precision=1e8):
    """  Returns values at the given quantile.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series. NOTE: the time series should be sorted.
    :param q: Percentile(s) at which to extract score(s). One or many.
    :param precision: Number of decimals expected.
    :return: Values at the given quantile.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.quantile_statistics(ctypes.pointer(tss.arr_reference),
                                                       ctypes.pointer(q.arr_reference),
                                                       ctypes.pointer(ctypes.c_float(precision)),
                                                       ctypes.pointer(b))
    return Array(array_reference=b)


def quantiles_cut(tss, quantiles, precision=1e-8):
    """ Discretizes the time series into equal-sized buckets based on sample quantiles.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series. NOTE: the time series should be sorted.
    :param quantiles: Number of quantiles to extract. From 0 to 1, step 1/quantiles.
    :param precision: Number of decimals expected.
    :return: Matrix with the categories, one category per row, the start of the category in the first column and
            the end in the second category.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.quantiles_cut_statistics(ctypes.pointer(tss.arr_reference),
                                                            ctypes.pointer(ctypes.c_float(quantiles)),
                                                            ctypes.pointer(ctypes.c_float(precision)),
                                                            ctypes.pointer(b))
    return Array(array_reference=b)


def sample_stdev(tss):
    """ Estimates standard deviation based on a sample. The standard deviation is calculated using the "n-1" method.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series.
    :return: The sample standard deviation.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.sample_stdev_statistics(ctypes.pointer(tss.arr_reference),
                                                           ctypes.pointer(b))
    return Array(array_reference=b)


def skewness(tss):
    """ Calculates the sample skewness of tss (calculated with the adjusted Fisher-Pearson standardized moment
    coefficient G1).

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series. NOTE: the time series should be sorted.
    :return: Array containing the skewness of each time series in tss.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.skewness_statistics(ctypes.pointer(tss.arr_reference),
                                                       ctypes.pointer(b))
    return Array(array_reference=b)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
import numpy as np
from khiva.normalization import *
from khiva.array import Array
from khiva.library import set_backend, KHIVABackend


########################################################################################################################

class NormalizationTest(unittest.TestCase):
    DELTA = 1e-6
    DECIMAL = 6

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_znorm(self):
        znorm_result = znorm(Array([[0, 1, 2, 3], [4, 5, 6, 7]]), 0.00000001).to_numpy().flatten()
        expected = [-1.341640786499870, -0.447213595499958, 0.447213595499958, 1.341640786499870]
        for i in range(len(expected)):
            self.assertAlmostEqual(znorm_result[i], expected[i], delta=self.DELTA)
            self.assertAlmostEqual(znorm_result[i + 4], expected[i], delta=self.DELTA)

    def test_znorm_in_place(self):
        tss = Array(data=[[0, 1, 2, 3], [4, 5, 6, 7]])
        znorm_in_place(tss)
        tss = tss.to_numpy()
        self.assertAlmostEqual(tss[0][0], -1.341640786499870, delta=self.DELTA)
        self.assertAlmostEqual(tss[0][1], -0.447213595499958, delta=self.DELTA)
        self.assertAlmostEqual(tss[0][2], 0.447213595499958, delta=self.DELTA)
        self.assertAlmostEqual(tss[0][3], 1.341640786499870, delta=self.DELTA)

        self.assertAlmostEqual(tss[1][0], -1.341640786499870, delta=self.DELTA)
        self.assertAlmostEqual(tss[1][1], -0.447213595499958, delta=self.DELTA)
        self.assertAlmostEqual(tss[1][2], 0.447213595499958, delta=self.DELTA)
        self.assertAlmostEqual(tss[1][3], 1.341640786499870, delta=self.DELTA)

    def test_max_min_norm(self):
        max_min_norm_result = max_min_norm(Array([[0, 1, 2, 3], [4, 5, 6, 7]]), 2.0, 1.0).to_numpy()
        expected = np.array([[1.0, 1.3333333333333, 1.66666667, 2.0], [1.0, 1.3333333333333, 1.66666667, 2.0]])
        np.testing.assert_array_almost_equal(max_min_norm_result, expected, decimal=self.DECIMAL)

    def test_max_min_norm_in_place(self):
        tss = Array([[0, 1, 2, 3], [4, 5, 6, 7]])
        max_min_norm_in_place(tss, 2.0, 1.0)
        tss = tss.to_numpy()
        expected = np.array([[1.0, 1.3333333333333, 1.66666667, 2.0], [1.0, 1.3333333333333, 1.66666667, 2.0]])
        np.testing.assert_array_almost_equal(tss, expected, decimal=self.DECIMAL)

    def test_decimal_scaling_norm(self):
        decimal_scaling_norm_result = decimal_scaling_norm(Array([[0, 1, -2, 3], [40, 50, 60, -70]])).to_numpy()
        expected = np.array([[0.0, 0.1, -0.2, 0.3], [0.4, 0.5, 0.6, -0.7]])
        np.testing.assert_array_almost_equal(decimal_scaling_norm_result, expected, decimal=self.DECIMAL)

    def test_decimal_scaling_norm_in_place(self):
        tss = Array([[0, 1, -2, 3], [40, 50, 60, -70]])
        decimal_scaling_norm_in_place(tss)
        tss = tss.to_numpy()
        expected = np.array([[0.0, 0.1, -0.2, 0.3], [0.4, 0.5, 0.6, -0.7]])
        np.testing.assert_array_almost_equal(tss, expected, decimal=self.DECIMAL)

    def test_mean_norm(self):
        result = mean_norm(Array([[0, 1, 2, 3], [4, 5, 6, 7]])).to_numpy()
        expected = np.array([[-0.5, -0.166666667, 0.166666667, 0.5], [-0.5, -0.166666667, 0.166666667, 0.5]])
        np.testing.assert_array_almost_equal(result, expected, decimal=self.DECIMAL)

    def test_mean_norm_in_place(self):
        a = Array([[0, 1, 2, 3], [4, 5, 6, 7]])
        mean_norm_in_place(a)
        expected = np.array([[-0.5, -0.166666667, 0.166666667, 0.5], [-0.5, -0.166666667, 0.166666667, 0.5]])
        np.testing.assert_array_almost_equal(a.to_numpy(), expected, decimal=self.DECIMAL)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(NormalizationTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest

import arrayfire as af
import numpy as np
import pandas as pd

from khiva.array import Array, dtype
from khiva.library import set_backend, KHIVABackend


########################################################################################################################

class ArrayTest(unittest.TestCase):
    DELTA = 1e-6
    DECIMAL = 6

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_real_1d_creation(self):
        a = Array([1, 5, 3, 1])
        np.testing.assert_array_equal(a.dims, np.array([4, 1, 1, 1]))

    def test_single_value_creation(self):
        a = Array([1])
        np.testing.assert_array_equal(a.dims, np.array([1, 1, 1, 1]))

    def test_real_1d(self):
        a = Array([1, 2, 3, 4, 5, 6, 7, 8])
        expected = np.array([1, 2, 3, 4, 5, 6, 7, 8])
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_real_2d_creation(self):
        a = Array([[1, 5, 3, 1], [2, 6, 9, 8], [3, 4, 1, 3]])
        np.testing.assert_array_equal(a.dims, np.array([4, 3, 1, 1]))

    def test_real_2d(self):
        a = Array([[1, 2, 3, 4], [5, 6, 7, 8]])
        expected = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_real_3d_creation(self):
        a = Array([[[1, 5, 3, 1], [2, 6, 9, 8], [3, 4, 1, 3]],
                   [[3, 7, 4, 2], [4, 8, 1, 9], [1, 5, 9, 2]]])
        np.testing.assert_array_equal(a.dims, np.array([4, 3, 2, 1]))

    def test_real_3d(self):
        a = Array([[[1, 5], [2, 6]], [[3, 7], [4, 8]]])
        expected = np.array([[[1, 5], [2, 6]], [[3, 7], [4, 8]]])
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_real_3d_large_column(self):
        a = Array([[[1, 5, 3], [2, 6, 9]], [[3, 7, 4], [4, 8, 1]]])
        expected = np.array([[[1, 5, 3], [2, 6, 9]], [[3, 7, 4], [4, 8, 1]]])
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_real_4d(self):
        a = Array([[[[1, 9], [2, 10]], [[3, 11], [4, 12]]], [[[5, 13], [6, 14]], [[7, 15], [8, 16]]]])
        expected = np.array([[[[1, 9], [2, 10]], [[3, 11], [4, 12]]], [[[5, 13], [6, 14]], [[7, 15], [8, 16]]]])
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex64_1d(self):
        a = Array(np.array([1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j]).astype(np.complex64), khiva_type=dtype.c32)
        expected = np.array([1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j]).astype(np.complex64)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex64_2d(self):
        a = Array(
            np.array([[1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j], [9 + 13j, 10 + 14j, 11 + 15j, 12 + 16j]]).astype(np.complex64),
            khiva_type=dtype.c32)
        expected = np.array([[1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j], [9 + 13j, 10 + 14j, 11 + 15j, 12 + 16j]]).astype(
            np.complex64)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex64_3d(self):
        a = Array(
            np.array([[[1 + 1j, 5 + 5j], [2 + 2j, 6 + 6j]], [[3 + 3j, 7 + 7j], [4 + 4j, 8 + 8j]]]).astype(np.complex64),
            khiva_type=dtype.c32)
        expected = np.array([[[1 + 1j, 5 + 5j], [2 + 2j, 6 + 6j]], [[3 + 3j, 7 + 7j], [4 + 4j, 8 + 8j]]]).astype(
            np.complex64)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex64_4d(self):
        a = Array(
            np.array([[[[1 + 1j, 9 + 9j], [2 + 2j, 10 + 10j]], [[3 + 3j, 11 + 11j], [4 + 4j, 12 + 12j]]],
                      [[[5 + 5j, 13 + 13j], [6 + 6j, 14 + 14j]], [[7 + 7j, 15 + 15j], [8 + 8j, 16 + 16j]]]])
                .astype(np.complex64),
            khiva_type=dtype.c32)
        expected = np.array([[[[1 + 1j, 9 + 9j], [2 + 2j, 10 + 10j]], [[3 + 3j, 11 + 11j], [4 + 4j, 12 + 12j]]],
                             [[[5 + 5j, 13 + 13j], [6 + 6j, 14 + 14j]],
                              [[7 + 7j, 15 + 15j], [8 + 8j, 16 + 16j]]]]).astype(
            np.complex64)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex128_1d(self):
        a = Array(np.array([1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j]).astype(np.complex128), khiva_type=dtype.c64)
        expected = np.array([1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j]).astype(np.complex128)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex128_2d(self):
        a = Array(
            np.array([[1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j], [9 + 13j, 10 + 14j, 11 + 15j, 12 + 16j]]).astype(np.complex128),
            khiva_type=dtype.c32)
        expected = np.array([[1 + 5j, 2 + 6j, 3 + 7j, 4 + 8j], [9 + 13j, 10 + 14j, 11 + 15j, 12 + 16j]]).astype(
            np.complex128)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex128_3d(self):
        a = Array(
            np.array([[[1 + 1j, 5 + 5j], [2 + 2j, 6 + 6j]], [[3 + 3j, 7 + 7j], [4 + 4j, 8 + 8j]]]).astype(
                np.complex128),
            khiva_type=dtype.c32)
        expected = np.array([[[1 + 1j, 5 + 5j], [2 + 2j, 6 + 6j]], [[3 + 3j, 7 + 7j], [4 + 4j, 8 + 8j]]]).astype(
            np.complex128)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_complex128_4d(self):
        a = Array(
            np.array([[[[1 + 1j, 9 + 9j], [2 + 2j, 10 + 10j]], [[3 + 3j, 11 + 11j], [4 + 4j, 12 + 12j]]],
                      [[[5 + 5j, 13 + 13j], [6 + 6j, 14 + 14j]], [[7 + 7j, 15 + 15j], [8 + 8j, 16 + 16j]]]])
                .astype(np.complex128),
            khiva_type=dtype.c32)
        expected = np.array([[[[1 + 1j, 9 + 9j], [2 + 2j, 10 + 10j]], [[3 + 3j, 11 + 11j], [4 + 4j, 12 + 12j]]],
                             [[[5 + 5j, 13 + 13j], [6 + 6j, 14 + 14j]],
                              [[7 + 7j, 15 + 15j], [8 + 8j, 16 + 16j]]]]).astype(
            np.complex128)
        np.testing.assert_array_equal(a.to_numpy(), expected)

    def test_get_type(self):
        a = Array([[1, 2, 3, 4], [5, 6, 7, 8]], khiva_type=dtype.s64)
        expected = dtype.s64
        self.assertEqual(a.get_type(), expected)

    def test_join(self):
        a = Array([1, 2, 3, 4], khiva_type=dtype.f64)
        b = Array([5, 6, 7, 8], khiva_type=dtype.f64)
        c = a.join(0, b)
        np.testing.assert_array_equal(c.to_numpy(), np.array([1, 2, 3, 4, 5, 6, 7, 8]))

    def testPlus(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        c = a + b
        np.testing.assert_array_equal(c.to_numpy(), np.array([2, 4, 6, 8]))

    def testTimes(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        c = a * b
        np.testing.assert_array_equal(c.to_numpy(), np.array([1, 4, 9, 16]))

    def testMinus(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        c = a - b
        np.testing.assert_array_equal(c.to_numpy(), np.array([0, 0, 0, 0]))

    def testRDivide(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        c = a / b
        expected = np.array([1., 1., 1., 1.])
        for d, e in zip(expected, c.to_numpy()):
            self.assertAlmostEqual(d, e, delta=1e-5)

    def testMod(self):
        a = Array([1, 2, 3, 4])
        b = Array([2, 2, 2, 2])
        c = a % b
        np.testing.assert_array_equal(c.to_numpy(), np.array([1, 0, 1, 0]))

    def testPower(self):
        a = Array([1, 2, 3, 4])
        b = Array([2, 2, 2, 2])
        c = a ** b
        np.testing.assert_array_equal(c.to_numpy(), np.array([1, 4, 9, 16]))

    def testLt(self):
        a = Array([1, 2, 3, 4])
        b = Array([2, 2, 2, 2])
        c = a < b
        np.testing.assert_array_equal(c.to_numpy(), np.array([True, False, False, False]))

    def testGt(self):
        a = Array([1, 2, 3, 4])
        b = Array([2, 2, 2, 2])
        c = a > b
        np.testing.assert_array_equal(c.to_numpy(), np.array([False, False, True, True]))

    def testLe(self):
        a = Array([1, 2, 3, 4])
        b = Array([2, 2, 2, 2])
        c = a <= b
        np.testing.assert_array_equal(c.to_numpy(), np.array([True, True, False, False]))

    def testGe(self):
        a = Array([1, 2, 3, 4])
        b = Array([2, 2, 2, 2])
        c = a >= b
        np.testing.assert_array_equal(c.to_numpy(), np.array([False, True, True, True]))

    def testEq(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 5])
        c = a == b
        np.testing.assert_array_equal(c.to_numpy(), np.array([True, True, True, False]))

    def testNeq(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 5])
        c = a != b
        np.testing.assert_array_equal(c.to_numpy(), np.array([False, False, False, True]))

    def testAnd(self):
        a = Array([True, True, True, True], dtype.b8)
        b = Array([True, False, True, False], dtype.b8)
        c = a & b
        np.testing.assert_array_equal(c.to_numpy(), np.array([True, False, True, False]))

    def testOr(self):
        a = Array([True, True, True, True], dtype.b8)
        b = Array([True, False, True, False], dtype.b8)
        c = a | b
        np.testing.assert_array_equal(c.to_numpy(), np.array([True, True, True, True]))

    def testXor(self):
        a = Array([True, True, True, True], dtype.b8)
        b = Array([True, False, True, False], dtype.b8)
        c = a ^ b
        np.testing.assert_array_equal(c.to_numpy(), np.array([False, True, False, True]))

    def testBitshift(self):
        a = Array([2, 4, 6, 8], dtype.s32)
        c = a >> 1
        np.testing.assert_array_equal(c.to_numpy(), np.array([1, 2, 3, 4]))

    def testBitsra(self):
        a = Array([2, 4, 6, 8], dtype.s32)
        c = a << 1
        np.testing.assert_array_equal(c.to_numpy(), np.array([4, 8, 12, 16]))

    def testCtranspose(self):
        a = Array([[0 - 1j, 4 + 2j], [2 + 1j, 0 - 2j]], khiva_type=dtype.c32)
        b = a.transpose(True)
        expected = [[0 + 1j, 2 - 1j], [4 - 2j, 0 + 2j]]
        np.testing.assert_array_equal(b.to_numpy(), expected)

    def testTranspose(self):
        a = Array([[1, 3], [2, 4]], dtype.s32)
        c = a.transpose()
        np.testing.assert_array_equal(c.to_numpy(), np.array([[1, 2], [3, 4]]))

    def testCol(self):
        a = Array([[1, 3], [2, 4]], dtype.s32)
        c = a.get_col(0)
        np.testing.assert_array_equal(c.to_numpy(), np.array([1, 3]))

    def testCols(self):
        a = Array(np.transpose([[1, 2, 3], [4, 5, 6]]), dtype.s32)
        c = a.get_cols(0, 1)
        np.testing.assert_array_equal(c.to_numpy(), np.transpose(np.array([[1, 2], [4, 5]])))

    def testRow(self):
        a = Array(np.transpose([[1, 2], [3, 4]]), dtype.s32)
        c = a.get_row(0)
        np.testing.assert_array_equal(c.to_numpy(), np.transpose(np.array([[1, 2]])))

    def testRows(self):
        a = Array(np.transpose([[1, 2], [3, 4], [5, 6]]), dtype.s32)
        c = a.get_rows(0, 1)
        np.testing.assert_array_equal(c.to_numpy(), np.transpose(np.array([[1, 2], [3, 4]])))

    def testMtimes(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        d = a.transpose()
        c = b.matmul(d)
        expected = np.transpose([[1, 2, 3, 4], [2, 4, 6, 8], [3, 6, 9, 12], [4, 8, 12, 16]])
        np.testing.assert_array_equal(c.to_numpy(), expected)

    def testAs(self):
        a = Array([1, 2, 3, 4], khiva_type=dtype.s32)
        b = a.as_type(dtype.u32)
        expected_data = [1, 2, 3, 4]
        np.testing.assert_array_equal(b.to_numpy(), expected_data)
        self.assertEqual(b.khiva_type, dtype.u32)

    def testCopy(self):
        a = Array([1, 2, 3, 4], khiva_type=dtype.s32)
        b = a.copy()
        np.testing.assert_array_equal(a.to_numpy(), b.to_numpy())
        self.assertEqual(b.khiva_type, b.khiva_type)

    def testArrayfire(self):
        a = af.Array([1, 2, 3, 4])
        a_data = a.to_list()
        b = Array.from_arrayfire(a)
        np.testing.assert_array_equal(np.asarray(a_data), np.asarray(b.to_list()))

    def testFromPandas(self):
        df = pd.DataFrame([[1, 2, 3, 4], [5, 6, 7, 8]])
        df_array = Array(df, khiva_type=dtype.s32)
        np.testing.assert_array_equal(df.values, df_array.to_pandas().values)

    def testLength(self):
        a = Array([1, 2, 3, 4])
        self.assertEqual(len(a), 4)

    def testIadd(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        a += b
        np.testing.assert_array_equal(a.to_numpy(), np.array([2, 4, 6, 8]))

    def testIaddSelfArray(self):
        a = Array([1, 2, 3, 4, 5])
        a += a
        np.testing.assert_array_equal(a.to_numpy(), np.array([2, 4, 6, 8, 10]))

    def testISub(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        a -= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([0, 0, 0, 0]))

    def testIMul(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        a *= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([1, 4, 9, 16]))

    def testITrueDiv(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        a /= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([1, 1, 1, 1]))

    def testDiv(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        c = a / b
        np.testing.assert_array_equal(c.to_numpy(), np.array([1, 1, 1, 1]))

    def testMod(self):
        a = Array([1, 2, 3, 4])
        b = Array([1, 2, 3, 4])
        a %= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([0, 0, 0, 0]))

    def testIPow(self):
        a = Array([1, 2, 3, 4])
        b = Array([2, 2, 2, 2])
        a **= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([1, 4, 9, 16]))

    def testIAnd(self):
        a = Array([1, 1, 1, 1], khiva_type=dtype.b8)
        b = Array([1, 0, 1, 0], khiva_type=dtype.b8)
        a &= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([1, 0, 1, 0]))

    def testIOr(self):
        a = Array([1, 1, 1, 1], khiva_type=dtype.b8)
        b = Array([1, 0, 1, 0], khiva_type=dtype.b8)
        a |= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([1, 1, 1, 1]))

    def testXor(self):
        a = Array([True, True, True, True], dtype.b8)
        b = Array([True, False, True, False], dtype.b8)
        a ^= b
        np.testing.assert_array_equal(a.to_numpy(), np.array([False, True, False, True]))

    def testBitshift(self):
        a = Array([2, 4, 6, 8], dtype.s32)
        a >>= 1
        np.testing.assert_array_equal(a.to_numpy(), np.array([1, 2, 3, 4]))

    def testBitsra(self):
        a = Array([2, 4, 6, 8], dtype.s32)
        a <<= 1
        np.testing.assert_array_equal(a.to_numpy(), np.array([4, 8, 12, 16]))

    def testNeg(self):
        a = Array([[1, 2], [3, 4]])
        b = -a
        np.testing.assert_array_equal(b.to_numpy(), np.array([[-1, -2], [-3, -4]]))

    def testNot(self):
        a = Array([True, True, True, False], khiva_type=dtype.b8)
        b = ~a
        np.testing.assert_array_equal(b.to_numpy(), np.array([False, False, False, True]))

    def testStr(self):
        a = Array([2, 4, 6, 8], dtype.s32)
        self.assertTrue("khiva.Array()\nType: dtype.s32\nDims: [4 1 1 1]" == str(a))

    def testRepre(self):
        a = Array([1, 2, 3, 4])
        self.assertTrue("khiva.Array()\nType: dtype.f32\nDims: [4 1 1 1]" == str(repr(a)))

    def testNonZero(self):
        a = Array([1, 2, 3, 4])
        self.assertTrue(a)

    def testGetData1(self):
        a = Array([1])
        np.testing.assert_array_equal(a._get_data(), np.array([1]))

    def testToList1(self):
        a = Array([1])
        self.assertIs(type(a.to_list()), list)
        self.assertEqual(a.to_list()[0], 1)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(ArrayTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################

def decimal_scaling_norm(tss):
    """ Normalizes the given time series according to its maximum value and adjusts each value within the range (-1, 1).

    :param tss: KHIVA array with the time series.

    :return: KHIVA array with the same dimensions as tss, whose values (time series in dimension 0) have been
             normalized by dividing each number by :math::`10^j` , where j is the number of integer digits of the max number in
             the time series.`
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.decimal_scaling_norm(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def decimal_scaling_norm_in_place(tss):
    """ Same as decimal_scaling_norm, but it performs the operation in place, without allocating further memory.

    :param tss: KHIVA array with the time series.
    """
    KhivaLibrary().c_khiva_library.decimal_scaling_norm_in_place(ctypes.pointer(tss.arr_reference))


def max_min_norm(tss, high=1.0, low=0.0, epsilon=0.00000001):
    """ Normalizes the given time series according to its minimum and maximum value and adjusts each value within the
    range [low, high].

    :param tss: KHIVA array with the time series.
    :param high: Maximum final value (Defaults to 1.0).
    :param low: Minimum final value (Defaults to 0.0).
    :param epsilon: Safeguard for constant (or near constant) time series as the operation implies a unit scale
                    operation between min and max values in the tss.

    :return: KHIVA array with the same dimensions as tss where the time series have been adjusted for zero mean and
            one as standard deviation.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.max_min_norm(ctypes.pointer(tss.arr_reference),
                                                ctypes.pointer(ctypes.c_double(high)),
                                                ctypes.pointer(ctypes.c_double(low)),
                                                ctypes.pointer(ctypes.c_double(epsilon)),
                                                ctypes.pointer(b))

    return Array(array_reference=b)


def max_min_norm_in_place(tss, high=1.0, low=0.0, epsilon=0.00000001):
    """ Same as max_min_norm, but it performs the operation in place, without allocating further memory.

    :param tss: KHIVA array with the time series.
    :param high: Maximum final value (Defaults to 1.0).
    :param low: Minimum final value (Defaults to 0.0).
    :param epsilon: Safeguard for constant (or near constant) time series as the operation implies a unit scale
                    operation between min and max values in the tss.
    """
    KhivaLibrary().c_khiva_library.max_min_norm_in_place(ctypes.pointer(tss.arr_reference),
                                                         ctypes.pointer(ctypes.c_double(high)),
                                                         ctypes.pointer(ctypes.c_double(low)),
                                                         ctypes.pointer(ctypes.c_double(epsilon)))


def mean_norm(tss):
    """ Normalizes the given time series according to its maximum-minimum value and its mean. It follows the following
    formulae:

    .. math::
        \\acute{x} = frac{x - mean(x)}{max(x) - min(x)}.

    :param tss: KHIVA array with the time series.

    :return: An array with the same dimensions as tss, whose values (time series in dimension 0) have been
            normalized by substracting the mean from each number and dividing each number by :math:`max(x) - min(`x)`,
            in the time series.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.mean_norm(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def mean_norm_in_place(tss):
    """ Normalizes the given time series according to its maximum-minimum value and its mean. It follows the following
    formulae:

    .. math:
        \\acute{x} = frac{x - mean(x)}{max(x) - min(x)}.

    :param tss: KHIVA array with the time series.
    """
    KhivaLibrary().c_khiva_library.mean_norm_in_place(ctypes.pointer(tss.arr_reference))


def znorm(tss, epsilon=0.00000001):
    """ Calculates a new set of time series with zero mean and standard deviation one.

    :param tss: KHIVA array with the time series.
    :param epsilon: Minimum standard deviation to consider. It acts as a gatekeeper for those time series that
           may be constant or near constant.

    :return: KHIVA array with the same dimensions as tss where the time series have been adjusted for zero mean and
            one as standard deviation.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.znorm(ctypes.pointer(tss.arr_reference), ctypes.pointer(ctypes.c_double(epsilon)),
                                         ctypes.pointer(b))

    return Array(array_reference=b)


def znorm_in_place(tss, epsilon=0.00000001):
    """ Adjusts the time series in the given input and performs z-norm
    in place (without allocating further memory).

    :param tss: KHIVA array with the time series.
    :param epsilon: epsilon Minimum standard deviation to consider. It acts as a gatekeeper for
                    those time series that may be constant or near constant.
    """
    KhivaLibrary().c_khiva_library.znorm_in_place(ctypes.pointer(tss.arr_reference),
                                                  ctypes.pointer(ctypes.c_double(epsilon)))



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################

def find_best_n_discords(profile, index, m, n, self_join=False):
    """ This function extracts the best N motifs from a previously calculated matrix profile.

    :param profile: KHIVA array with the matrix profile containing the minimum distance of each subsequence.
    :param index: KHIVA array with the matrix profile index containing where each minimum occurs.
    :param m: Subsequence length value used to calculate the input matrix profile.
    :param n: Number of discords to extract.
    :param self_join: Indicates whether the input profile comes from a self join operation or not. It determines
                      whether the mirror similar region is included in the output or not.
    :return: KHIVA arrays with the discord distances, the discord indices and the subsequence indices.
    """
    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)
    d = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.find_best_n_discords(ctypes.pointer(profile.arr_reference),
                                                        ctypes.pointer(index.arr_reference),
                                                        ctypes.pointer(ctypes.c_long(m)),
                                                        ctypes.pointer(ctypes.c_long(n)),
                                                        ctypes.pointer(b),
                                                        ctypes.pointer(c),
                                                        ctypes.pointer(d),
                                                        ctypes.pointer(ctypes.c_bool(self_join)))

    return Array(array_reference=b), Array(array_reference=c), Array(
        array_reference=d)


def find_best_n_motifs(profile, index, m, n, self_join=False):
    """ This function extracts the best N discords from a previously calculated matrix profile.

    :param profile: KHIVA array with the matrix profile containing the minimum distance of each subsequence.
    :param index: KHIVA array with the matrix profile index containing where each minimum occurs.
    :param m: Subsequence length value used to calculate the input matrix profile.
    :param n: Number of motifs to extract.
    :param self_join: Indicates whether the input profile comes from a self join operation or not. It determines
                      whether the mirror similar region is included in the output or not.
    :return: KHIVA arrays with the motif distances, the motif indices and the subsequence indices.
    """
    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)
    d = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.find_best_n_motifs(ctypes.pointer(profile.arr_reference),
                                                      ctypes.pointer(index.arr_reference),
                                                      ctypes.pointer(ctypes.c_long(m)),
                                                      ctypes.pointer(ctypes.c_long(n)),
                                                      ctypes.pointer(b),
                                                      ctypes.pointer(c),
                                                      ctypes.pointer(d),
                                                      ctypes.pointer(ctypes.c_bool(self_join)))

    return Array(array_reference=b), Array(array_reference=c), Array(
        array_reference=d)


def find_best_n_occurrences(query_time_series, time_series, number_of_occurrences):
    """ Calculates the N best matches of several queries in several time series.

     The result has the following structure:
        - 1st dimension corresponds to the nth best match.
        - 2nd dimension corresponds to the number of queries.
        - 3rd dimension corresponds to the number of time series.

    For example, the distance in the position (1, 2, 3) corresponds to the second best distance of the third query in the
    fourth time series. The index in the position (1, 2, 3) is the index of the subsequence which leads to the
    second best distance of the third query in the fourth time series.

    :param query_time_series: Array whose first dimension is the length of the query time series and the second
    dimension is the number of queries.
    :param time_series: Array whose first dimension is the length of the time series and the second dimension is the
    number of time series.
    :param number_of_occurrences: Number of matches to return.
    :return: KHIVA arrays with the distances and indexes.
    """

    distances = ctypes.c_void_p(0)
    indexes = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.find_best_n_occurrences(ctypes.pointer(query_time_series.arr_reference),
                                                           ctypes.pointer(time_series.arr_reference),
                                                           ctypes.pointer(ctypes.c_long(number_of_occurrences)),
                                                           ctypes.pointer(distances),
                                                           ctypes.pointer(indexes))

    return Array(array_reference=distances), Array(array_reference=indexes)


def mass(query_time_series, time_series):
    """ Mueen's Algorithm for Similarity Search.

     The result has the following structure:
        - 1st dimension corresponds to the index of the subsequence in the time series.
        - 2nd dimension corresponds to the number of queries.
        - 3rd dimension corresponds to the number of time series.

    For example, the distance in the position (1, 2, 3) correspond to the distance of the third query to the fourth time
    series for the second subsequence in the time series.

    :param query_time_series: Array whose first dimension is the length of the query time series and the second
    dimension is the number of queries.
    :param time_series: Array whose first dimension is the length of the time series and the second dimension is the
    number of time series.
    :return: KHIVA array with the distances.
    """

    distances = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.mass(ctypes.pointer(query_time_series.arr_reference),
                                                           ctypes.pointer(time_series.arr_reference),
                                                           ctypes.pointer(distances))

    return Array(array_reference=distances)


def stomp(first_time_series, second_time_series, subsequence_length):
    """ Stomp algorithm to calculate the matrix profile between `ta` and `tb` using a subsequence length of `m`.

    [1] Yan Zhu, Zachary Zimmerman, Nader Shakibay Senobari, Chin-Chia Michael Yeh, Gareth Funning, Abdullah Mueen,
    Philip Brisk and Eamonn Keogh (2016). Matrix Profile II: Exploiting a Novel Algorithm and GPUs to break the one
    Hundred Million Barrier for Time Series Motifs and Joins. IEEE ICDM 2016.

    :param first_time_series: KHIVA array with the first time series.
    :param second_time_series: KHIVA array with the second time series.
    :param subsequence_length: Length of the subsequence.
    :return: KHIVA arrays with the profile and index.
    """

    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.stomp(ctypes.pointer(first_time_series.arr_reference),
                                         ctypes.pointer(second_time_series.arr_reference),
                                         ctypes.pointer(ctypes.c_long(subsequence_length)),
                                         ctypes.pointer(b),
                                         ctypes.pointer(c))

    return Array(array_reference=b), Array(array_reference=c)


def stomp_self_join(time_series, subsequence_length):
    """ Stomp algorithm to calculate the matrix profile between `t` and itself using a subsequence length of `m`.
    This method filters the trivial matches.

    [1] Yan Zhu, Zachary Zimmerman, Nader Shakibay Senobari, Chin-Chia Michael Yeh, Gareth Funning, Abdullah Mueen,
    Philip Brisk and Eamonn Keogh (2016). Matrix Profile II: Exploiting a Novel Algorithm and GPUs to break the one
    Hundred Million Barrier for Time Series Motifs and Joins. IEEE ICDM 2016.

    :param time_series: The query and reference time series in KHIVA array format.
    :param subsequence_length: Lenght of the subsequence
    :return: KHIVA arrays with the profile and index.
    """
    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.stomp_self_join(ctypes.pointer(time_series.arr_reference),
                                                   ctypes.pointer(ctypes.c_long(subsequence_length)),
                                                   ctypes.pointer(b),
                                                   ctypes.pointer(c))

    return Array(array_reference=b), Array(array_reference=c)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
import numpy as np
from khiva.features import *
from khiva.array import Array, dtype
from khiva.library import set_backend, KHIVABackend
import arrayfire as af


########################################################################################################################

class FeaturesTest(unittest.TestCase):
    DELTA = 1e-6

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_cid_ce(self):
        cid_ce_result = cid_ce(Array(data=[[0, 1, 2, 3, 4, 5], [6, 7, 8, 9, 10, 11]]),
                               False).to_numpy()
        self.assertAlmostEqual(cid_ce_result[0], 2.23606797749979, delta=self.DELTA)
        self.assertAlmostEqual(cid_ce_result[1], 2.23606797749979, delta=self.DELTA)

        cid_ce_result = cid_ce(Array(data=[[0, 1, 2, 3, 4, 5], [6, 7, 8, 9, 10, 11]]),
                               True).to_numpy()
        self.assertAlmostEqual(cid_ce_result[0], 1.30930734141595, delta=self.DELTA)
        self.assertAlmostEqual(cid_ce_result[1], 1.30930734141595, delta=self.DELTA)

    def test_c3(self):
        c3_result = c3(Array(data=[[0, 1, 2, 3, 4, 5], [6, 7, 8, 9, 10, 11]]), 2).to_numpy()
        self.assertEqual(c3_result[0], 7.5)
        self.assertEqual(c3_result[1], 586.5)

    def test_abs_sum_of_changes(self):
        abs_sum_of_changes_result = absolute_sum_of_changes(
            Array(data=[[0, 1, 2, 3], [4, 6, 8, 10], [11, 14, 17, 20]])).to_numpy()
        self.assertEqual(abs_sum_of_changes_result[0], 3)
        self.assertEqual(abs_sum_of_changes_result[1], 6)
        self.assertEqual(abs_sum_of_changes_result[2], 9)

    def test_abs_energy(self):
        abs_energy_result = abs_energy(Array(data=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10])).to_numpy()
        self.assertAlmostEqual(abs_energy_result, 385, delta=self.DELTA)

    def test_cross_correlation(self):
        cross_correlation_result = cross_correlation(xss=Array(data=[1, 2, 3, 4]),
                                                     yss=Array(data=[4, 6, 8, 10, 12]),
                                                     unbiased=False).to_numpy()
        self.assertAlmostEqual(cross_correlation_result[0], 0.790569415, delta=self.DELTA)
        self.assertAlmostEqual(cross_correlation_result[1], 0.790569415, delta=self.DELTA)
        self.assertAlmostEqual(cross_correlation_result[2], 0.079056941, delta=self.DELTA)
        self.assertAlmostEqual(cross_correlation_result[3], -0.395284707, delta=self.DELTA)
        self.assertAlmostEqual(cross_correlation_result[4], -0.474341649, delta=self.DELTA)

    def test_auto_covariance(self):
        auto_covariance_result = auto_covariance(
            Array(data=[[0, 1, 2, 3], [10, 11, 12, 13]])).to_numpy()
        self.assertAlmostEqual(auto_covariance_result[0][0], 1.25, delta=self.DELTA)
        self.assertAlmostEqual(auto_covariance_result[0][1], 0.3125, delta=self.DELTA)
        self.assertAlmostEqual(auto_covariance_result[0][2], -0.375, delta=self.DELTA)
        self.assertAlmostEqual(auto_covariance_result[0][3], -0.5625, delta=self.DELTA)
        self.assertAlmostEqual(auto_covariance_result[1][0], 1.25, delta=self.DELTA)
        self.assertAlmostEqual(auto_covariance_result[1][1], 0.3125, delta=self.DELTA)
        self.assertAlmostEqual(auto_covariance_result[1][2], -0.375, delta=self.DELTA)
        self.assertAlmostEqual(auto_covariance_result[1][3], -0.5625, delta=self.DELTA)

    def test_cross_covariance(self):
        cross_covariance_result = cross_covariance(
            xss=Array(data=[[0, 1, 2, 3], [10, 11, 12, 13]]),
            yss=Array(data=[[4, 6, 8, 10, 12], [14, 16, 18, 20, 22]]),
            unbiased=False).to_numpy().flatten()
        for i in range(4):
            self.assertAlmostEqual(cross_covariance_result[(i * 5)], 2.5, delta=self.DELTA)
            self.assertAlmostEqual(cross_covariance_result[(i * 5) + 1], 2.5, delta=self.DELTA)
            self.assertAlmostEqual(cross_covariance_result[(i * 5) + 2], 0.25, delta=self.DELTA)
            self.assertAlmostEqual(cross_covariance_result[(i * 5) + 3], -1.25, delta=self.DELTA)
            self.assertAlmostEqual(cross_covariance_result[(i * 5) + 4], -1.5, delta=self.DELTA)

    def test_approximate_entropy(self):
        approximate_entropy_result = approximate_entropy(
            Array(data=[[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], [11, 12, 13, 14, 15, 16, 17, 18, 19, 20]],
                  khiva_type=dtype.f32), 4, 0.5).to_numpy()

        self.assertAlmostEqual(approximate_entropy_result[0], 0.13484281753639338, delta=self.DELTA)
        self.assertAlmostEqual(approximate_entropy_result[1], 0.13484281753639338, delta=self.DELTA)

    def test_auto_correlation(self):
        auto_correlation_result = auto_correlation(
            Array(data=[[0, 1, 2, 3], [10, 11, 12, 13]]), 4, False).to_numpy().flatten()
        self.assertAlmostEqual(auto_correlation_result[0], 1, delta=self.DELTA)
        self.assertAlmostEqual(auto_correlation_result[1], 0.25, delta=self.DELTA)
        self.assertAlmostEqual(auto_correlation_result[2], -0.3, delta=self.DELTA)
        self.assertAlmostEqual(auto_correlation_result[3], -0.45, delta=self.DELTA)
        self.assertAlmostEqual(auto_correlation_result[4], 1.0, delta=self.DELTA)
        self.assertAlmostEqual(auto_correlation_result[5], 0.25, delta=self.DELTA)
        self.assertAlmostEqual(auto_correlation_result[6], -0.3, delta=self.DELTA)
        self.assertAlmostEqual(auto_correlation_result[7], -0.45, delta=self.DELTA)

    def test_binned_entropy(self):
        binned_entropy_result = binned_entropy(Array(data=[[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,
                                                            14, 15, 16, 17, 18, 19, 20],
                                                           [1, 1, 3, 10, 5, 6, 1, 8, 9, 10, 11, 1, 13, 14, 10, 16,
                                                            17, 10, 19,
                                                            20]]), 5).to_numpy()
        self.assertAlmostEqual(binned_entropy_result[0], 1.6094379124341005, delta=self.DELTA)
        self.assertAlmostEqual(binned_entropy_result[1], 1.5614694247763998, delta=self.DELTA)

    def test_count_above_mean(self):
        count_above_mean_result = count_above_mean(Array(data=[[0, 1, 2, 3, 4, 5],
                                                               [6, 7, 8, 9, 10, 11]])).to_numpy()
        self.assertAlmostEqual(count_above_mean_result[0], 3, delta=self.DELTA)
        self.assertAlmostEqual(count_above_mean_result[1], 3, delta=self.DELTA)

    def test_count_below_mean(self):
        count_below_mean_result = count_below_mean(Array(data=[[0, 1, 2, 3, 4, 5],
                                                               [6, 7, 8, 9, 10, 11]])).to_numpy()
        self.assertAlmostEqual(count_below_mean_result[0], 3, delta=self.DELTA)
        self.assertAlmostEqual(count_below_mean_result[1], 3, delta=self.DELTA)

    def test_energy_ratio_by_chunks(self):
        energy_ratio_by_chunks_result = energy_ratio_by_chunks(Array([[0, 1, 2, 3, 4, 5],
                                                                      [6, 7, 8, 9, 10, 11]]),
                                                               2, 0).to_numpy()
        self.assertAlmostEqual(energy_ratio_by_chunks_result[0], 0.090909091, delta=self.DELTA)
        self.assertAlmostEqual(energy_ratio_by_chunks_result[1], 0.330376940, delta=self.DELTA)
        energy_ratio_by_chunks_result = energy_ratio_by_chunks(Array([[0, 1, 2, 3, 4, 5],
                                                                      [6, 7, 8, 9, 10, 11]]), 2,
                                                               1).to_numpy()
        self.assertAlmostEqual(energy_ratio_by_chunks_result[0], 0.909090909, delta=self.DELTA)
        self.assertAlmostEqual(energy_ratio_by_chunks_result[1], 0.669623060, delta=self.DELTA)

    def test_first_location_of_maximum(self):
        first_location_of_maximum_result = first_location_of_maximum(
            Array([[5, 4, 3, 5, 0, 1, 5, 3, 2, 1], [2, 4, 3, 5, 2, 5, 4, 3, 5, 2]])).to_numpy()
        self.assertAlmostEqual(first_location_of_maximum_result[0], 0.0, delta=self.DELTA)
        self.assertAlmostEqual(first_location_of_maximum_result[1], 0.3, delta=self.DELTA)

    def test_first_location_of_minimum(self):
        first_location_of_minimum_result = first_location_of_minimum(
            Array([[5, 4, 3, 0, 0, 1], [5, 4, 3, 0, 2, 1]])).to_numpy()
        self.assertAlmostEqual(first_location_of_minimum_result[0], 0.5, delta=self.DELTA)
        self.assertAlmostEqual(first_location_of_minimum_result[1], 0.5, delta=self.DELTA)

    def test_friedrich_coefficients(self):
        friedrich_coefficients_result = friedrich_coefficients(
            Array([[0, 1, 2, 3, 4, 5], [0, 1, 2, 3, 4, 5]]), 4, 2).to_numpy()
        expected = np.array([[-0.0009912563255056738, -0.0027067768387496471, -0.00015192681166809052,
                              0.10512571036815643, 0.89872437715530396],
                             [-0.0009912563255056738, -0.0027067768387496471, -0.00015192681166809052,
                              0.10512571036815643, 0.89872437715530396]])
        np.testing.assert_array_almost_equal(friedrich_coefficients_result, expected, decimal=6)

    def test_has_duplicates(self):
        has_duplicates_result = has_duplicates(
            Array([[5, 4, 3, 0, 0, 1], [5, 4, 3, 0, 2, 1]])).to_numpy()
        self.assertEqual(has_duplicates_result[0], True)
        self.assertEqual(has_duplicates_result[1], False)

    def test_has_duplicate_max(self):
        has_duplicate_max_result = has_duplicate_max(
            Array([[5, 4, 3, 0, 5, 1], [5, 4, 3, 0, 2, 1]])).to_numpy()
        self.assertEqual(has_duplicate_max_result[0], True)
        self.assertEqual(has_duplicate_max_result[1], False)

    def test_index_mass_quantile(self):
        index_mass_quantile_result = index_mass_quantile(Array([[5, 4, 3, 0, 0, 1], [5, 4, 0, 0, 2, 1]]),
                                                         0.5).to_numpy()

        self.assertAlmostEqual(index_mass_quantile_result[0], 0.333333333, delta=self.DELTA)
        self.assertAlmostEqual(index_mass_quantile_result[1], 0.333333333, delta=self.DELTA)

    def test_kurtosis(self):
        kurtosis_result = kurtosis(Array([[0, 1, 2, 3, 4, 5], [2, 2, 2, 20, 30, 25]])).to_numpy()
        self.assertAlmostEqual(kurtosis_result[0], -1.2, delta=1e-4)
        self.assertAlmostEqual(kurtosis_result[1], -2.66226722, delta=1e-4)

    def test_large_standard_deviation(self):
        large_standard_deviation_result = large_standard_deviation(
            Array([[-1, -1, -1, 1, 1, 1], [4, 6, 8, 4, 5, 4]]), 0.4).to_numpy()
        self.assertEqual(large_standard_deviation_result[0], True)
        self.assertEqual(large_standard_deviation_result[1], False)

    def test_last_location_of_maximum(self):
        last_location_of_maximum_result = last_location_of_maximum(
            Array([[0, 4, 3, 5, 5, 1], [0, 4, 3, 2, 5, 1]])).to_numpy()
        self.assertAlmostEqual(last_location_of_maximum_result[0], 0.8333333333333334, delta=self.DELTA)
        self.assertAlmostEqual(last_location_of_maximum_result[1], 0.8333333333333334, delta=self.DELTA)

    def test_last_location_of_minimum(self):
        last_location_of_minimum_result = last_location_of_minimum(
            Array([[0, 4, 3, 5, 5, 1, 0, 4], [3, 2, 5, 1, 4, 5, 1, 2]])).to_numpy()
        self.assertAlmostEqual(last_location_of_minimum_result[0], 0.875, delta=self.DELTA)
        self.assertAlmostEqual(last_location_of_minimum_result[1], 0.875, delta=self.DELTA)

    def test_length(self):
        length_result = length(Array([[0, 4, 3, 5, 5, 1], [0, 4, 3, 2, 5, 1]])).to_numpy().flatten()
        self.assertEqual(length_result[0], 6)
        self.assertEqual(length_result[1], 6)

    def test_linear_trend(self):
        pvalue, rvalue, intercept, slope, stderr = linear_trend(Array([[0, 4, 3, 5, 5, 1], [2, 4, 1, 2, 5, 3]]))

        pvalue = pvalue.to_numpy()
        self.assertAlmostEqual(pvalue[0], 0.6260380997892747, delta=self.DELTA)
        self.assertAlmostEqual(pvalue[1], 0.5272201945463578, delta=self.DELTA)
        rvalue = rvalue.to_numpy()
        self.assertAlmostEqual(rvalue[0], 0.2548235957188128, delta=self.DELTA)
        self.assertAlmostEqual(rvalue[1], 0.3268228676411533, delta=self.DELTA)
        intercept = intercept.to_numpy()
        self.assertAlmostEqual(intercept[0], 2.2857142857142856, delta=self.DELTA)
        self.assertAlmostEqual(intercept[1], 2.1904761904761907, delta=self.DELTA)
        slope = slope.to_numpy()
        self.assertAlmostEqual(slope[0], 0.2857142857142857, delta=self.DELTA)
        self.assertAlmostEqual(slope[1], 0.2571428571428572, delta=self.DELTA)
        stderr = stderr.to_numpy()
        self.assertAlmostEqual(stderr[0], 0.5421047417431507, delta=self.DELTA)
        self.assertAlmostEqual(stderr[1], 0.37179469135129783, delta=self.DELTA)

    def test_has_duplicate_min(self):
        has_duplicate_min_result = has_duplicate_min(
            Array(data=[[5, 4, 3, 0, 0, 1], [5, 4, 3, 0, 2, 1]])).to_numpy()
        self.assertEqual(has_duplicate_min_result[0], True)
        self.assertEqual(has_duplicate_min_result[1], False)

    def test_longest_strike_above_mean(self):
        longest_strike_above_mean_result = longest_strike_above_mean(
            Array(data=[[20, 20, 20, 1, 1, 1, 20, 20, 20, 20, 1, 1, 1, 1,
                         1, 1, 1, 1, 20, 20],
                        [20, 20, 20, 1, 1, 1, 20, 20,
                         20, 1, 1, 1, 1, 1, 1, 1, 1,
                         1, 20, 20]])).to_numpy()
        self.assertEqual(longest_strike_above_mean_result[0], 4)
        self.assertEqual(longest_strike_above_mean_result[1], 3)

    def test_longest_strike_below_mean(self):
        longest_strike_below_mean_result = longest_strike_below_mean(
            Array([[20, 20, 20, 1, 1, 1, 20, 20, 20, 20, 1, 1, 1, 1,
                    1, 1, 1, 1, 20, 20],
                   [20, 20, 20, 1, 1, 1, 20, 20,
                    20, 1, 1, 1, 1, 1, 1, 1, 1,
                    1, 20, 20]], dtype.f32)).to_numpy()
        self.assertEqual(longest_strike_below_mean_result[0], 8)
        self.assertEqual(longest_strike_below_mean_result[1], 9)

    def test_maximum(self):
        maximum_result = maximum(Array([[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 1, 1, 5, 1, 20, 20],
                                        [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20,
                                         20]])).to_numpy()
        self.assertEqual(maximum_result[0], 50)
        self.assertEqual(maximum_result[1], 30)

    def test_mean_absolute_change(self):
        mean_absolute_change_result = mean_absolute_change(
            Array([[0, 1, 2, 3, 4, 5], [8, 10, 12, 14, 16, 18]])).to_numpy()
        r = 5.0 / 6.0
        self.assertAlmostEqual(mean_absolute_change_result[0], r, delta=self.DELTA)
        self.assertAlmostEqual(mean_absolute_change_result[1], r * 2, delta=self.DELTA)

    def test_fft_coefficient(self):
        fftCoefficient_result = fft_coefficient(
            Array([[0, 1, 2, 3, 4, 5], [6, 7, 8, 9, 10, 11]]), 0)
        a = fftCoefficient_result[0].to_numpy()
        self.assertAlmostEqual(a[0], 15, delta=self.DELTA)
        self.assertAlmostEqual(a[1], 51, delta=self.DELTA)
        b = fftCoefficient_result[1].to_numpy()

        self.assertAlmostEqual(b[0], 0, delta=self.DELTA)
        self.assertAlmostEqual(b[1], 0, delta=self.DELTA)
        c = fftCoefficient_result[2].to_numpy()

        self.assertAlmostEqual(c[0], 15, delta=self.DELTA)
        self.assertAlmostEqual(c[1], 51, delta=self.DELTA)
        d = fftCoefficient_result[3].to_numpy()

        self.assertAlmostEqual(d[0], 0, delta=self.DELTA)
        self.assertAlmostEqual(d[1], 0, delta=self.DELTA)

    def test_aggregated_autocorrelation_mean(self):
        aggregated_autocorrelation_result = aggregated_autocorrelation(
            Array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]]), 0).to_numpy()
        self.assertAlmostEqual(aggregated_autocorrelation_result[0], -0.6571428571428571, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_autocorrelation_result[1], -0.6571428571428571, delta=self.DELTA)

    def test_aggregated_autocorrelation_median(self):
        aggregated_autocorrelation_result = aggregated_autocorrelation(
            Array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]]), 1).to_numpy()
        self.assertAlmostEqual(aggregated_autocorrelation_result[0], -0.54285717010498047, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_autocorrelation_result[1], -0.54285717010498047, delta=self.DELTA)

    def test_aggregated_autocorrelation_min(self):
        aggregated_autocorrelation_result = aggregated_autocorrelation(
            Array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]]), 2).to_numpy()
        self.assertAlmostEqual(aggregated_autocorrelation_result[0], -2.142857142857143, delta=1e-4)
        self.assertAlmostEqual(aggregated_autocorrelation_result[1], -2.142857142857143, delta=1e-4)

    def test_aggregated_autocorrelation_max(self):
        aggregated_autocorrelation_result = aggregated_autocorrelation(
            Array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]]), 3).to_numpy()
        self.assertAlmostEqual(aggregated_autocorrelation_result[0], 0.6, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_autocorrelation_result[1], 0.6, delta=self.DELTA)

    def test_aggregated_autocorrelation_stdev(self):
        aggregated_autocorrelation_result = aggregated_autocorrelation(
            Array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]]), 4).to_numpy()
        self.assertAlmostEqual(aggregated_autocorrelation_result[0], 0.9744490855905009, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_autocorrelation_result[1], 0.9744490855905009, delta=self.DELTA)

    def test_aggregated_autocorrelation_var(self):
        aggregated_autocorrelation_result = aggregated_autocorrelation(
            Array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]]), 5).to_numpy()
        self.assertAlmostEqual(aggregated_autocorrelation_result[0], 0.9495510204081633, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_autocorrelation_result[1], 0.9495510204081633, delta=self.DELTA)

    def test_aggregated_linear_trend_mean(self):
        aggregated_linear_trend_result = aggregated_linear_trend(
            Array([[2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5]]), 3, 0)
        self.assertAlmostEqual(aggregated_linear_trend_result[0].to_numpy(), 1, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_linear_trend_result[1].to_numpy(), 2, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_linear_trend_result[2].to_numpy(), 1, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_linear_trend_result[3].to_numpy(), 0, delta=self.DELTA)
        self.assertAlmostEqual(aggregated_linear_trend_result[4].to_numpy(), 0, delta=self.DELTA)

    def test_aggregated_linear_trend_min(self):
        aggregated_linear_trend_result = aggregated_linear_trend(
            Array([2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5]), 3, 2)
        self.assertAlmostEqual(aggregated_linear_trend_result[0].to_numpy(), 1, delta=1e-3)
        self.assertAlmostEqual(aggregated_linear_trend_result[1].to_numpy(), 2, delta=1e-3)
        self.assertAlmostEqual(aggregated_linear_trend_result[2].to_numpy(), 1, delta=1e-3)
        self.assertAlmostEqual(aggregated_linear_trend_result[3].to_numpy(), 0, delta=1e-3)
        self.assertAlmostEqual(aggregated_linear_trend_result[4].to_numpy(), 0, delta=1e-3)

    def test_cwt_coefficients(self):
        cwt_coefficients_result = cwt_coefficients(Array([[0.1, 0.2, 0.3], [0.1, 0.2, 0.3]]),
                                                   Array(data=[1, 2, 3], khiva_type=dtype.s32), 2, 2).to_numpy()
        self.assertAlmostEqual(cwt_coefficients_result[0], 0.26517161726951599, delta=self.DELTA)
        self.assertAlmostEqual(cwt_coefficients_result[1], 0.26517161726951599, delta=self.DELTA)

    def test_mean_second_derivative_central(self):
        mean_second_derivative_central_result = mean_second_derivative_central(
            Array([[1, 3, 7, 4, 8], [2, 5, 1, 7, 4]])).to_numpy()
        self.assertAlmostEqual(mean_second_derivative_central_result[0], 1.0 / 5.0, delta=self.DELTA)
        self.assertAlmostEqual(mean_second_derivative_central_result[1], -3.0 / 5.0, delta=self.DELTA)

    def test_minimum(self):
        minimum_result = minimum(
            Array([[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 13, 15, 5, 16, 20, 20],
                   [20, 20, 20, 2, 19, 4, 20, 20, 20, 4, 15, 6, 30, 7, 9, 18, 4, 10, 20, 20]])).to_numpy()
        self.assertAlmostEqual(minimum_result[0], 1, delta=self.DELTA)
        self.assertAlmostEqual(minimum_result[1], 2, delta=self.DELTA)

    def test_number_crossing_m(self):
        number_crossing_m_result = number_crossing_m(
            Array([[1, 2, 1, 1, -3, -4, 7, 8, 9, 10, -2, 1, -3, 5, 6, 7, -10],
                   [1, 2, 1, 1, -3, -4, 7, 8, 9, 10, -2, 1, -3, 5, 6, 7, -10]]), 0).to_numpy()

        self.assertAlmostEqual(number_crossing_m_result[0], 7, delta=self.DELTA)
        self.assertAlmostEqual(number_crossing_m_result[1], 7, delta=self.DELTA)

    def test_mean(self):
        mean_result = mean(Array([[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 1, 1, 5, 1, 20, 20],
                                  [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20, 20]])).to_numpy()
        self.assertAlmostEqual(mean_result[0], 18.55, delta=1e-4)
        self.assertAlmostEqual(mean_result[1], 12.7, delta=1e-4)

    def test_median(self):
        median_result = median(Array([[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 1, 1, 5, 1, 20, 20],
                                      [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20,
                                       20]])).to_numpy()
        self.assertAlmostEqual(median_result[0], 20, delta=self.DELTA)
        self.assertAlmostEqual(median_result[1], 18.5, delta=self.DELTA)

    def test_mean_change(self):
        mean_change_result = mean_change(Array([[0, 1, 2, 3, 4, 5],
                                                [8, 10, 12, 14, 16, 18]])).to_numpy()
        self.assertAlmostEqual(mean_change_result[0], 5.0 / 6.0, delta=self.DELTA)
        self.assertAlmostEqual(mean_change_result[1], 10.0 / 6.0, delta=self.DELTA)

    def test_max_langevin_fixed_point(self):
        max_langevin_fixed_point_result = max_langevin_fixed_point(
            Array([[0, 1, 2, 3, 4, 5], [0, 1, 2, 3, 4, 5]]), 7, 2).to_numpy()
        self.assertAlmostEqual(max_langevin_fixed_point_result[0], 4.562970585, delta=1e-4)
        self.assertAlmostEqual(max_langevin_fixed_point_result[1], 4.562970585, delta=1e-4)

    def test_fft_aggregated(self):
        fft_aggregated_result = fft_aggregated(
            Array([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]],
                  dtype.f32)).to_numpy().flatten()
        self.assertAlmostEqual(fft_aggregated_result[0], 1.135143, delta=1e-4)
        self.assertAlmostEqual(fft_aggregated_result[1], 2.368324, delta=1e-4)
        self.assertAlmostEqual(fft_aggregated_result[2], 1.248777, delta=1e-4)
        self.assertAlmostEqual(fft_aggregated_result[3], 3.642666, delta=1e-4)

        self.assertAlmostEqual(fft_aggregated_result[4], 1.135143, delta=1e-4)
        self.assertAlmostEqual(fft_aggregated_result[5], 2.368324, delta=1e-4)
        self.assertAlmostEqual(fft_aggregated_result[6], 1.248777, delta=1e-4)
        self.assertAlmostEqual(fft_aggregated_result[7], 3.642666, delta=1e-4)

    def test_number_cwt_peaks(self):
        result = number_cwt_peaks(Array([[1, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1],
                                         [1, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 1, 1, 1]]),
                                  2).to_numpy()
        self.assertEqual(result[0], 2)
        self.assertEqual(result[1], 2)

    def test_number_peaks(self):
        result = number_peaks(Array([[3, 0, 0, 4, 0, 0, 13], [3, 0, 0, 4, 0, 0, 13]]), 2).to_numpy()
        self.assertAlmostEqual(result[0], 1, delta=1e-4)
        self.assertAlmostEqual(result[1], 1, delta=1e-4)

    def test_partial_autocorrelation(self):
        numel = 3000.0
        step = 1 / (numel - 1)
        a = [step * i for i in range(int(numel))]
        a = Array([a, a])
        lags = Array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], khiva_type=dtype.s32)
        result = partial_autocorrelation(a, lags).to_numpy()
        expected = np.array([[1.0, 0.9993331432342529, -0.0006701064994559, -0.0006701068487018, -0.0008041285327636,
                              -0.0005360860959627, -0.0007371186511591, -0.0004690756904893, -0.0008041299879551,
                              -0.0007371196406893],
                             [1.0, 0.9993331432342529, -0.0006701064994559, -0.0006701068487018, -0.0008041285327636,
                              -0.0005360860959627, -0.0007371186511591, -0.0004690756904893, -0.0008041299879551,
                              -0.0007371196406893]])
        np.testing.assert_array_almost_equal(result, expected, decimal=3)

    def test_percentage_of_reocurring_datapoints_to_all_datapoints(self):
        result = percentage_of_reoccurring_datapoints_to_all_datapoints(
            Array([[3, 0, 0, 4, 0, 0, 13], [3, 0, 0, 4, 0, 0, 13]]), False).to_numpy()
        self.assertAlmostEqual(result[0], 0.25, delta=1e-4)
        self.assertAlmostEqual(result[1], 0.25, delta=1e-4)

    def test_percentage_of_reocurring_values_to_all_values(self):
        result = percentage_of_reoccurring_values_to_all_values(
            Array([[1, 1, 2, 3, 4, 4, 5, 6], [1, 2, 2, 3, 4, 5, 6, 7]]), False).to_numpy()
        self.assertEqual(result[0], 4.0 / 8.0)
        self.assertEqual(result[1], 2.0 / 8.0)

    def test_quantile(self):
        result = quantile(Array([[0, 0, 0, 0, 3, 4, 13], [0, 0, 0, 0, 3, 4, 13]]),
                          Array([0.6], dtype.f32)).to_numpy()
        self.assertAlmostEqual(result[0], 1.79999999, delta=1e-4)
        self.assertAlmostEqual(result[1], 1.79999999, delta=1e-4)

    def test_range_count(self):
        result = range_count(Array([[3, 0, 0, 4, 0, 0, 13], [3, 0, 5, 4, 0, 0, 13]]), 2, 12).to_numpy()
        self.assertEqual(result[0], 2)
        self.assertEqual(result[1], 3)

    def test_ratio_beyond_r_sigma(self):
        result = ratio_beyond_r_sigma(Array([[3, 0, 0, 4, 0, 0, 13], [3, 0, 0, 4, 0, 0, 13]]),
                                      0.5).to_numpy()
        self.assertAlmostEqual(result[0], 0.7142857142857143, delta=1e-4)
        self.assertAlmostEqual(result[1], 0.7142857142857143, delta=1e-4)

    def test_ratio_value_number_to_time_series_length(self):
        result = ratio_value_number_to_time_series_length(
            Array([[3, 0, 0, 4, 0, 0, 13], [3, 5, 0, 4, 6, 0, 13]])).to_numpy()
        self.assertAlmostEqual(result[0], 4.0 / 7.0, delta=1e-4)
        self.assertAlmostEqual(result[1], 6.0 / 7.0, delta=1e-4)

    def test_sample_entropy(self):
        result = sample_entropy(Array([[3, 0, 0, 4, 0, 0, 13], [3, 0, 0, 4, 0, 0, 13]])).to_numpy()
        self.assertAlmostEqual(result[0], 1.2527629, delta=1e-4)
        self.assertAlmostEqual(result[1], 1.2527629, delta=1e-4)

    def test_skewness(self):
        result = skewness(Array([[3, 0, 0, 4, 0, 0, 13], [3, 0, 0, 4, 0, 0, 13]])).to_numpy()
        self.assertAlmostEqual(result[0], 2.038404735373753, delta=1e-4)
        self.assertAlmostEqual(result[1], 2.038404735373753, delta=1e-4)

    def test_spkt_welch_density(self):
        result = spkt_welch_density(Array([[0, 1, 1, 3, 4, 5, 6, 7, 8, 9], [0, 1, 1, 3, 4, 5, 6, 7, 8, 9]]),
                                    0).to_numpy()
        self.assertAlmostEqual(result[0], 1.6666667, delta=1e-5)
        self.assertAlmostEqual(result[1], 1.6666667, delta=1e-5)

    def test_standard_deviation(self):
        result = standard_deviation(
            Array([[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 1, 1, 5, 1, 20, 20],
                   [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20, 20]])).to_numpy()
        self.assertAlmostEqual(result[0], 12.363150892875165, delta=1e-4)
        self.assertAlmostEqual(result[1], 9.51367436903324, delta=1e-4)

    def test_sum_of_reoccurring_datapoints(self):
        result = sum_of_reoccurring_datapoints(Array([[3, 3, 0, 4, 0, 13, 13], [3, 3, 0, 4, 0, 13, 13]])).to_numpy()
        self.assertEqual(result[0], 32)
        self.assertEqual(result[1], 32)

    def test_sum_of_reoccurring_values(self):
        result = sum_of_reoccurring_values(Array([[4, 4, 6, 6, 7], [4, 7, 7, 8, 8]])).to_numpy()
        self.assertEqual(result[0], 10)
        self.assertEqual(result[1], 15)

    def test_sum_values(self):
        result = sum_values(Array([[1, 2, 3, 4.1], [-1.2, -2, -3, -4]])).to_numpy()
        self.assertAlmostEqual(result[0], 10.1, delta=self.DELTA)
        self.assertAlmostEqual(result[1], -10.2, delta=self.DELTA)

    def test_symmetry_looking(self):
        result = symmetry_looking(Array([[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 1, 1, 5, 1, 20, 20],
                                         [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20, 20]]),
                                  0.1).to_numpy()
        self.assertEqual(result[0], bool(1))
        self.assertEqual(result[1], bool(0))

    def test_time_reversal_asymmetry_statistic(self):
        result = time_reversal_asymmetry_statistic(Array(
            [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20],
             [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20, 20]]), 2).to_numpy()
        self.assertEqual(result[0], 1052)
        self.assertEqual(result[1], -150.625)

    def test_value_count(self):
        result = value_count(Array(data=[[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 1, 1, 5, 1, 20, 20],
                                         [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20, 20]]),
                             20).to_numpy()
        self.assertAlmostEqual(result[0], 9, delta=1e-4)
        self.assertAlmostEqual(result[1], 8, delta=1e-4)

    def test_variance(self):
        result = variance(Array(data=[[1, 1, -1, -1], [1, 2, -2, -1]])).to_numpy()
        self.assertEqual(result[0], 1)
        self.assertEqual(result[1], 2.5)

    def test_variance_larger_than_standard_deviation(self):
        result = variance_larger_than_standard_deviation(
            Array(data=[[20, 20, 20, 18, 25, 19, 20, 20, 20, 20, 40, 30, 1, 50, 1, 1, 5, 1, 20, 20],
                        [20, 20, 20, 2, 19, 1, 20, 20, 20, 1, 15, 1, 30, 1, 1, 18, 4, 1, 20, 20]])).to_numpy()
        self.assertEqual(result[0], True)
        self.assertEqual(result[1], True)

    def test_concatenated(self):
        a = Array([[1, 2], [1, 3], [1, 4], [1, 5], [1, 6], [1, 7], [1, 8], [1, 9], [1, 10], [1, 11]])
        b = absolute_sum_of_changes(a).to_arrayfire()
        c = af.transpose(b)
        d = Array.from_arrayfire(c)
        e = abs_energy(d).to_numpy()
        self.assertAlmostEqual(e, 385, delta=self.DELTA)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(FeaturesTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################

def group_by(tss, aggregation_function, n_columns_key=1, n_columns_value=1):
    """ Group by operation in the input array using n_columns_key columns as group keys and n_columns_value columns as
    values. The data is expected to be sorted. The aggregation function determines the operation to aggregate the
    values.

    :param tss: KHIVA array with the time series.
    :param aggregation_function: Function to be used in the aggregation. It receives an integer which indicates
                                the function to be applied.
                                0 : mean,
                                1 : median
                                2 : min,
                                3 : max,
                                4 : stdev,
                                5 : var,
                                default : mean
    :param n_columns_key: Number of columns conforming the key.
    :param n_columns_value: Number of columns conforming the value (they are expected to be consecutive to the column
                            keys).

    :return: KHIVA array with the values of the group keys aggregated using the aggregation_function.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.group_by(ctypes.pointer(tss.arr_reference),
                                            ctypes.pointer(ctypes.c_int(aggregation_function)),
                                            ctypes.pointer(ctypes.c_int(n_columns_key)),
                                            ctypes.pointer(ctypes.c_int(n_columns_value)),
                                            ctypes.pointer(b))

    return Array(array_reference=b)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
from khiva.distances import *
from khiva.array import Array
import numpy as np
from khiva.library import set_backend, KHIVABackend


########################################################################################################################


class DistancesTest(unittest.TestCase):

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_euclidean(self):
        euclidean_result = euclidean(Array(data=[[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11]])).to_numpy().flatten()
        expected = np.array([0, 0, 0, 8, 0, 0, 16, 8, 0])
        np.testing.assert_array_almost_equal(euclidean_result, expected, decimal=1)

    def test_dtw(self):
        euclidean_result = dtw(Array(
            data=[[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3], [4, 4, 4, 4, 4], [5, 5, 5, 5, 5]])).to_numpy()
        expected = np.array([[0, 0, 0, 0, 0], [5, 0, 0, 0, 0], [10, 5, 0, 0, 0], [15, 10, 5, 0, 0], [20, 15, 10, 5, 0]])
        np.testing.assert_array_almost_equal(euclidean_result, expected, decimal=1)

    def test_hamming(self):
        result = hamming(Array(
            data=[[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3], [4, 4, 4, 4, 4], [5, 5, 5, 5, 5]])).to_numpy()
        expected = np.array([[0, 0, 0, 0, 0], [5, 0, 0, 0, 0], [5, 5, 0, 0, 0], [5, 5, 5, 0, 0], [5, 5, 5, 5, 0]])
        np.testing.assert_array_almost_equal(result, expected, decimal=1)

    def test_manhattan(self):
        result = manhattan(Array(
            data=[[1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, 3, 3], [4, 4, 4, 4, 4], [5, 5, 5, 5, 5]])).to_numpy()
        expected = np.array([[0, 0, 0, 0, 0], [5, 0, 0, 0, 0], [10, 5, 0, 0, 0], [15, 10, 5, 0, 0], [20, 15, 10, 5, 0]])
        np.testing.assert_array_almost_equal(result, expected, decimal=2)

    def test_sbd(self):
        sbd_result = sbd(
            Array(data=[[1, 2, 3, 4, 5], [1, 1, 0, 1, 1], [10, 12, 0, 0, 1]])).to_numpy().flatten()
        expected = np.array([0, 0, 0, 0.505025, 0, 0, 0.458583, 0.564093, 0])
        np.testing.assert_array_almost_equal(sbd_result, expected, decimal=1)

    def test_squared_euclidean(self):
        squared_euclidean_result = squared_euclidean(
            Array(data=[[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11]])).to_numpy().flatten()
        expected = np.array([0, 0, 0, 64, 0, 0, 256, 64, 0])
        np.testing.assert_array_almost_equal(squared_euclidean_result, expected, decimal=1)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(DistancesTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################

def k_means(tss, k, tolerance=1e-10, max_iterations=100):
    """ Calculates the K-Means algorithm.

    [1] S. Lloyd. 1982. Least squares quantization in PCM. IEEE Transactions on Information Theory, 28, 2,
    Pages 129-137.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
    dimension one indicates the number of time series.
    :param k:                   The number of means to be computed.
    :param tolerance:           The error tolerance to stop the computation of the centroids.
    :param max_iterations:      The maximum number of iterations allowed.

    :return: Tuple with an array of centroids and array of labels.
    """
    centroids = ctypes.c_void_p(0)
    labels = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.k_means(ctypes.pointer(tss.arr_reference),
                                           ctypes.pointer(ctypes.c_int(k)),
                                           ctypes.pointer(centroids),
                                           ctypes.pointer(labels),
                                           ctypes.pointer(ctypes.c_float(tolerance)),
                                           ctypes.pointer(ctypes.c_int(max_iterations))
                                           )
    return Array(array_reference=centroids), Array(array_reference=labels)


def k_shape(tss, k, tolerance=1e-10, max_iterations=100):
    """ Calculates the K-Shape algorithm.

    [1] John Paparrizos and Luis Gravano. 2016. k-Shape: Efficient and Accurate Clustering of Time Series.
    SIGMOD Rec. 45, 1 (June 2016), 69-76.


    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
    dimension one indicates the number of time series.
    :param k:                   The number of means to be computed.
    :param tolerance:           The error tolerance to stop the computation of the centroids.
    :param max_iterations:      The maximum number of iterations allowed.

    :return: Tuple with an array of centroids and array of labels.
    """
    centroids = ctypes.c_void_p(0)
    labels = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.k_shape(ctypes.pointer(tss.arr_reference),
                                                          ctypes.pointer(ctypes.c_int(k)),
                                                          ctypes.pointer(centroids),
                                                          ctypes.pointer(labels),
                                                          ctypes.pointer(ctypes.c_float(tolerance)),
                                                          ctypes.pointer(ctypes.c_int(max_iterations))
                                                          )
    return Array(array_reference=centroids), Array(array_reference=labels)





#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from khiva.array import *
from khiva.clustering import *
from khiva.dimensionality import *
from khiva.distances import *
from khiva.features import *
from khiva.library import *
from khiva.linalg import *
from khiva.matrix import *
from khiva.normalization import *
from khiva.normalization import *
from khiva.polynomial import *
from khiva.regression import *
from khiva.regularization import *
from khiva.statistics import *



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
from enum import Enum
import ctypes
import platform
import logging
import sys


########################################################################################################################

class KhivaLibrary(object):
    class __KhivaLibrary:
        def __init__(self):
            try:
                if platform.system() == 'Darwin':
                    self.c_khiva_library = ctypes.CDLL('libkhiva_c.dylib')
                elif platform.system() == 'Windows':
                    self.c_khiva_library = ctypes.CDLL('C:/Program Files/Khiva/v0/lib/khiva_c.dll')
                elif platform.system() == 'Linux':
                    self.c_khiva_library = ctypes.CDLL('libkhiva_c.so')
            except:
                logging.error(
                    "Khiva C++ library is required in order to use the Python Khiva library")
                sys.exit(1)

    instance = None

    def __new__(cls):
        if not KhivaLibrary.instance:
            KhivaLibrary.instance = KhivaLibrary.__KhivaLibrary()
        return KhivaLibrary.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name):
        return setattr(self.instance, name)


class KHIVABackend(Enum):
    """
    KHIVA Backend.
    """
    KHIVA_BACKEND_DEFAULT = 0
    """
    Default Backend.
    """
    KHIVA_BACKEND_CPU = 1
    """
    CPU Backend.
    """
    KHIVA_BACKEND_CUDA = 2
    """
    CUDA Backend.
    """
    KHIVA_BACKEND_OPENCL = 4
    """
    OPENCL Backend.
    """


def get_backend_info():
    """ Gets information from the current backend.

    :return: A string with information from the current backend.
    """
    info_pointer = ctypes.c_char_p((" " * 1000).encode('utf8'))
    KhivaLibrary().c_khiva_library.backend_info(ctypes.pointer(info_pointer))
    return info_pointer.value.decode('utf8')


def set_backend(backend):
    """ Sets the KHIVABackend.

    :param backend: The desired backend. KHIVABackend type.
    """
    KhivaLibrary().c_khiva_library.set_backend(ctypes.pointer(ctypes.c_int(backend.value)))


def get_backend():
    """ Gets the active backend.

    :return: The active backend. KHIVABackend type.
    """
    backend = (ctypes.c_int * 1)(*[0])
    KhivaLibrary().c_khiva_library.get_backend(ctypes.pointer(backend))

    return KHIVABackend(backend[0])


def get_backends():
    """ Gets the available backends.

    :return: The available backends.
    """
    backends = (ctypes.c_int * 1)(*[0])
    KhivaLibrary().c_khiva_library.get_backends(ctypes.pointer(backends))
    return backends[0]


def set_device(device):
    """ Sets the device.

    :param device: The desired device.
    """
    KhivaLibrary().c_khiva_library.set_device(ctypes.pointer(ctypes.c_int(device)))


def get_device_id():
    """ Gets the active device.

    :return: The active device.
    """
    device = (ctypes.c_int * 1)(*[0])
    KhivaLibrary().c_khiva_library.get_device_id(ctypes.pointer(device))
    return device[0]


def get_device_count():
    """ Gets the devices count.

    :return: The devices count.
    """
    device_count = (ctypes.c_int * 1)(*[0])
    KhivaLibrary().c_khiva_library.get_device_count(ctypes.pointer(device_count))
    return device_count[0]


def version():
    """ Returns a string with the current version of the library.

    :return: A string with the current version of the library.
    """
    v = ctypes.c_char_p((" " * 40).encode('utf8'))
    KhivaLibrary().c_khiva_library.version(ctypes.pointer(v))
    return v.value.decode('utf8')



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################

def polyfit(x, y, deg):
    """ Least squares polynomial fit. Fit a polynomial :math:`p(x) = p[0] * x^{deg} + ... + p[deg]` of degree
    :math:`deg` to points :math:`(x, y)`. Returns a vector of coefficients :math:`p` that minimises the squared error.

    :param x: KHIVA array with the x-coordinates of the M sample points :math:`(x[i], y[i])`.
    :param y: KHIVA array with the y-coordinates of the sample points.
    :param deg: Degree of the fitting polynomial

    :return: KHIVA array with the polynomial coefficients, highest power first.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.polyfit(ctypes.pointer(x.arr_reference), ctypes.pointer(y.arr_reference),
                                           ctypes.pointer(ctypes.c_int(deg)),
                                           ctypes.pointer(b))

    return Array(array_reference=b)


def roots(p):
    """ Calculates the roots of a polynomial with coefficients given in :math:`p`. The values in the rank-1 array
    :math:`p` are coefficients of a polynomial. If the length of :math:`p` is :math:`n+1` then the polynomial is
    described by:

    .. math::
        p[0] * x^n + p[1] * x^{n-1} + ... + p[n-1] * x + p[n]

    :param p: KHIVA array with the polynomial coefficients.

    :return: KHIVA array with the roots of the polynomial.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.roots(ctypes.pointer(p.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################


def euclidean(tss):
    """ Calculates euclidean distances between time series.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
                dimension one indicates the number of time series.
    :return: Array with an upper triangular matrix where each position corresponds to the distance between two
            time series. Diagonal elements will be zero. For example: Position row 0 column 1 records the distance
            between time series 0 and time series 1.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.euclidean(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))
    return Array(array_reference=b)


def dtw(tss):
    """ Calculates the Dynamic Time Warping Distance.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
                dimension one indicates the number of time series.
    :return: Array with an upper triangular matrix where each position corresponds to the distance between
            two time series. Diagonal elements will be zero. For example: Position row 0 column 1 records the
            distance between time series 0 and time series 1.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.dtw(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))
    return Array(array_reference=b)


def hamming(tss):
    """ Calculates Hamming distances between time series.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
                dimension one indicates the number of time series.
    :return: Array with an upper triangular matrix where each position corresponds to the distance between two
            time series. Diagonal elements will be zero. For example: Position row 0 column 1 records the distance
            between time series 0 and time series 1.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.hamming(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))
    return Array(array_reference=b)


def manhattan(tss):
    """ Calculates Manhattan distances between time series.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
                dimension one indicates the number of time series.
    :return: Array with an upper triangular matrix where each position corresponds to the distance between two
            time series. Diagonal elements will be zero. For example: Position row 0 column 1 records the distance
            between time series 0 and time series 1.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.manhattan(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))
    return Array(array_reference=b)


def sbd(tss):
    """ Calculates the Shape-Based distance (SBD). It computes the normalized cross-correlation and
    it returns the value that maximizes the correlation value between time series.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
                dimension one indicates the number of time series.
    :return: Array with an upper triangular matrix where each position corresponds to the distance between two time series.
            Diagonal elements will be zero. For example: Position row 0 column 1 records the distance between time series 0
            and time series 1.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.sbd(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))
    return Array(array_reference=b)


def squared_euclidean(tss):
    """ Calculates the non squared version of the euclidean distance.

    :param tss: Expects an input array whose dimension zero is the length of the time series (all the same) and
                dimension one indicates the number of time series.
    :return: Array with an upper triangular matrix where each position corresponds to the distance between two time series.
            Diagonal elements will be zero. For example: Position row 0 column 1 records the distance between time series 0
            and time series 1.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.squared_euclidean(ctypes.pointer(tss.arr_reference), ctypes.pointer(b))
    return Array(array_reference=b)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
import numpy as np
from khiva.polynomial import *
from khiva.array import Array
from khiva.library import set_backend, KHIVABackend


########################################################################################################################

class PolynomialTest(unittest.TestCase):
    DELTA = 1e-6
    DECIMAL = 6

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_polyfit1(self):
        polyfit_result = polyfit(Array([0, 1, 2, 3, 4, 5]), Array([0, 1, 2, 3, 4, 5]), 1).to_numpy()
        expected = np.array([1.0, 0.0])
        np.testing.assert_array_almost_equal(polyfit_result, expected, decimal=self.DECIMAL)

    def test_polyfit3(self):
        polyfit_result = polyfit(Array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0]), Array([0.0, 0.8, 0.9, 0.1, -0.8, -1.0]),
                                 3).to_numpy()
        expected = np.array([0.08703704, -0.81349206, 1.69312169, -0.03968254])
        np.testing.assert_array_almost_equal(polyfit_result, expected, decimal=self.DECIMAL - 1)

    def test_roots(self):
        roots_result = roots(Array([5, -20, 5, 50, -20, -40])).to_numpy()
        expected = np.array([2 + 0j, 2 + 0j, 2 + 0j, -1 + 0j, -1 + 0j])
        np.testing.assert_array_almost_equal(roots_result, expected, decimal=2)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(PolynomialTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
